% A script to run a prototypal version of the Time Domain Vector Fitting 
% code (TDVF) on Single-Input Single-Output (SISO) test cases. 
%
% Data included in this package for testing the TDVF code refer to a
% cardiovascular modeling application, as described in section 3.5 of
%-------------------------------------------------------------------------
% Fevola, Elisa, et al. "A Vector Fitting approach for the automated
% estimation of lumped boundary conditions of 1D circulation models." 
% https://doi.org/10.48550/arXiv.2208.11482
%-----------------------------------------------------------------------
%
% The TDVF code is based on the real-time implementation introduced in
% -------------------------------------------------------------------
% Bradde, T.; Chevalier, S.; De Stefano, M.; Grivet-Talocia, S.; Daniel, L.
% Handling Initial Conditions in Vector Fitting for Real Time Modeling
% of Power System Dynamics. Energies 2021, 14, 2471.
% https://doi.org/10.3390/en14092471
% --------------------------------------------------------------------
%
% Notice that this implementation uses the matlab built-in function LSIM to
% perform the required convolution operations. More efficient
% implementations are possible, see 
% -------------------------------------------------------------------------
% S. Grivet-Talocia, "Package macromodeling via time-domain vector
% fitting," in IEEE Microwave and Wireless Components Letters, vol. 13, no.
% 11, pp. 472-474, Nov. 2003.
% https://doi.org/10.1109/LMWC.2003.819378
% -------------------------------------------------------------------------
%
% ----------------------------------------------
% Authors: Tommaso Bradde, Stefano Grivet-Talocia
% EMC group, Politecnico di Torino, Italy
% March 2023
%-----------------------------------------------

clear 
close all
clc

addpath('Data')
addpath('fun')

% load data. The training input u is the blood flow. The training output y
% is the pressure.
load('Segment3')

% define model order
Order = 3;

% define training signals
u = u(:);
y = y(:);
time = time(:);

% This redefinition would allow to use only part of the initial data for
% fitting the model
ufit=u(1:end);
yfit=y(1:end);
tfit=time(1:end);

% When this is set to zero the function will display no information
Options.Silent = 0;

% Normalize the training signals
[uNorm, yNorm, uConst, yConst] = signalsNormalization(ufit,yfit);

% Convergence condition: when the norm of the residues of the VF weighting
% function reaches this number, the VF iteration stops and the model
% residues are computed. Note that a more effective stopping condition can
% be implemented by checking when the (L2 or RMS) norm of the model/data
% error drops below some prescribed threshold. The latter choice is more
% robust in case of noisy data.
Options.VFWeightNorm = 1e-9;

% The maximum number of allowed iterations
Options.MaxIterations = 100;

% Option to enforce stability of the model
Options.EnforceStability = 1;

% Generate the model
[Model,weightingFunction,polesOut] = tdvfSISO(tfit,uNorm,yNorm,Order,Options);

% Denormalize the model and compute the error
[errors,ModH, yTemp] = DenormalizeAndValidateModel(Model,time,u,y,uConst,yConst);

%% This section below refers specifically to the application discussed in the cited preprint

% Compute the distal pressure by averaging the steady state error
% (this is here defined manually)
initSteady=9591;
endSteady=11502;
distalPressure=mean(errors.errorSignal(initSteady:endSteady));
yModel=yTemp+distalPressure;

% Plot the results
figure;
plot(time,y./133.33, 'linewidth',3, 'color', 'blue');

hold on
plot(time,yModel./133.33,  'linewidth',3, 'color', 'red');

legend('Data', 'Model');
title("Order " + Order);
xlabel('Time (s)');
ylabel('Pressure (mmHg)');
set(gca,'xlim',[10.463 11.522],'FontSize',14);

% Monitor convergence
figure;
semilogy(weightingFunction,'linewidth',2);
xlabel('Iterations');
ylabel('Norm of weighting function residues');
title('TDVF Convergence')
set(gca,'FontSize',14);
