function [errors,ModH,ymodel] = DenormalizeAndValidateModel(Realization,t,u,y,uConst,yConst)
%
% [errors,ModH,ymodel] = DenormalizeAndValidateModel(Realization,t,u,y,uConst,yConst)
%
% This function to be used in conjunction with signalsNormalization to
% denormalize model realization based on the same normalization constants
% used during model training.
%
% Inputs:
%
% Realization:   structure with fields A,B,C,D providing a normalized
%                state-space realization obtained by TDVF iterations
% t,u,v:         time, input and output signals to be used for model
%                validation (raw not normalized)
% uConst,yConst: normalization constants from signalsNormalization
%
% Outputs:
%
% errors:   structure with various time-domain error metrics
%           (self-explanatory) 
% ModH:     state-space object system of model (denormalized)
% ymodel:   output of the model excited by the validation input
%
% ----------------------------------------------
% Authors: Tommaso Bradde, Stefano Grivet-Talocia
% EMC group, Politecnico di Torino, Italy
% March 2023
%-----------------------------------------------

% Condition input signals
t = t(:);
u = u(:);
y = y(:);

% Remove "initial conditions" (first point, see
% https://doi.org/10.3390/en14092471) 
u = u-u(1);
y = y-y(1);

% This code valid also for the MIMO case (not provided in this package)
UU =  diag(1./uConst);
YY = diag(1./yConst);

% Denormalize model realization
AA = Realization.A;
BB = Realization.B*UU;
CC = YY\Realization.C;
DD = YY\Realization.D*UU;

% Use LSIM to compute model output
ModH=ss(AA,BB,CC,DD);
[ymodel,~,~] = lsim(ModH,u,t);

%% Compute errors
errors.errorSignal=y-ymodel;
errors.abs = abs(errors.errorSignal);
RMSvalue = sqrt(mean(y.^2,1));
errors.rel = errors.abs./RMSvalue;
errors.absMax = max(errors.abs);
errors.relMax = max(errors.rel);

return
end
