function [Realization,WeightFunctRes,polesOut] = tdvfSISO(t,u,y,Order,Options)
%
% [Realization,WeightFunctRes,polesOut] = tdvfSISO(t,u,y,Order,Options)
%
% A prototypal Time-Domain Vector Fitting code for SISO systems.
%
% The code implements the "real time" version of the TDVF algorithm, as
% described in
% -------------------------------------------------------------------
% Bradde, T.; Chevalier, S.; De Stefano, M.; Grivet-Talocia, S.; Daniel, L.
% Handling Initial Conditions in Vector Fitting for Real Time Modeling
% of Power System Dynamics. Energies 2021, 14, 2471.
% https://doi.org/10.3390/en14092471
% --------------------------------------------------------------------
%
% The code uses only standard built-in matlab functions (function lsim from
% Control Toolbox required) and is not optimized to efficiently compute
% convolutions. Details about a faster implementation can be found in
% -------------------------------------------------------------------------
% S. Grivet-Talocia, "Package macromodeling via time-domain vector
% fitting," in IEEE Microwave and Wireless Components Letters, vol. 13, no.
% 11, pp. 472-474, Nov. 2003.
% https://doi.org/10.1109/LMWC.2003.819378
% -------------------------------------------------------------------------
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%% INPUT ARGUMENTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% t: time vector. Must have the same number of samples as the input and
% output signals
%
% u,y: input and output signals used to train the model. NOTE: only
%      SISO systems are supported by this version of the code
%
% Order: model order (number of poles)
%
% Options: structure with control options for the TDVF algorithm. The
%          various fields are listed below (if missing, defaults are used)
%
% Options.Silent: when set to true, the code will provide no output logs
%
% Options.VFWeightNorm: stopping threshold for TDVF iterations. When the
% norm of the model denominator coefficients (residues of the VF weight
% function) falls below this value, the algorithm stops.
%
% Options.MaxIterations: maximum allowed number of iterations
%
% Options.EnforceStability: when set to true, the model poles will be
% forced to be stable. Since model is constructed in time-domain, this
% options must not be disabled (for frequency-domain VF it is possible to
% disable stability enforcement, at least during the first iterations)
%
% Options.tolFreqMax: option to control placement of intiial VF poles. This
% parameter must be in (0,1) and is used to reduce the initial poles
% bandwidth with respect to the maximum allowed bandwidth defined by the
% Nyquist rule. In particular, given a sampling frequency Fs (1/sampling
% time), the maximum frequency (in Hz) of the initial VF poles will be
% FreqMax=tolFreqMax*Fs/2;
%
% Options.FreqMin: lower band limit (in Hz) for intial VF poles placement.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%% OUTPUT ARGUMENTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Realization: model realization in state space form (structure with fields
% A,B,C,D)
%
% WeightFunctRes: norm of the denominator partial fractions residues
% vector at each VF iteration (used to monitor VF convergence)
%
% polesOut: the history of the model poles during the VF iterations
%
% ----------------------------------------------
% Authors: Tommaso Bradde, Stefano Grivet-Talocia
% EMC group, Politecnico di Torino, Italy
% March 2023
%-----------------------------------------------

%% Check and set algorithm options

% Set default options
def.Silent = 1;   % Silent mode disabled by default
def.debug = 0;   % debug info (poles)

% Stopping crition will be met if the norm of the residues of the VF
% weighting function is less than this value
def.VFWeightNorm= 1E-7;

% Maximum number of iterations
def.MaxIterations = 100;

% Skip identification of residues (computed only at the last iteration)
def.SkipRes = 1;

% Stability enforcement
def.EnforceStability = 1;

% Fraction of the maximum allowed frequency used for poles initialitation
% (linearly spaced complex poles in conjugate pairs)
def.tolFreqMax = 0.0001;

% Minimum frequency (in Hz) used for poles initialitation
def.FreqMin = 0.000001; % Hz

if nargin<5
    % Assign default options if missing on input
    Options=def;
else
    % Merge default values into Options
    A=fieldnames(def);
    for m=1:length(A)
        if ~isfield(Options,A{m})
            Options.(A{m}) = def.(A{m});
        end
    end
end

%% Basic initialization and checks

% Initialize outputs
Realization = [];
WeightFunctRes = ones(Options.MaxIterations,1);
polesOut=zeros(Order,Options.MaxIterations);

% Fix and check input arguments
if length(t) ~= numel(t) || length(u) ~= numel(u) || length(y) ~= numel(y)
    error('This code supports only SISO systems');
end
if length(t) ~= length(u) || length(u) ~= length(y)
    error('Input and output data should be column vectors with the same length of t.');
end
t = t(:);
u = u(:);
y = y(:);
kmax = length(y);

% Define time step from time axis vector and detect uniform sampling
% NOTE: uniform time samplint is required here, but it is not a fundamental
% prerequisite for TDVF.
[dT,errorDT] = extractTimeStep(t);

if ~isempty(errorDT)
    error(errorDT);
end

% Remove any initial conditions
u = u-u(1);
y = y-y(1);

%% Initialize starting VF poles

% Initial poles will be linearly spaced between these two angular
% frequencies
Fs = 1/dT;
Fmax = Fs/2;
omegamax = (2*pi*Fmax)*Options.tolFreqMax;
omegamin = 2*pi*Options.FreqMin;

% Number of real poles (no real poles if order is even)
nr = mod(Order,2); % 1 real pole if order is odd

% Number of complex conjugate pairs
nc = floor(Order/2);

% Basis poles placement...
% ... real poles
if nr == 1
    realpoles = -.5*(omegamax+omegamin);
else
    realpoles = [];
end
% ... complex poles
if nc >0
    complpoles = (-0.01+1j)*((omegamin*ones(1,nc))+((omegamax-omegamin)*linspace(0.5/nc,1-0.5/nc,nc)));
else
    complpoles = [];
end


%% TDVF core algorithm

% Current Iteration
iter = 0;

% Flag for poles convergence
PolesConverged=0;

if Options.debug
    fprintf(['\n Starting Poles:\n Real:' num2str(realpoles.') ' \n Complex: ' num2str(complpoles.') '\n']);
end

% Run VF iterations
while iter < Options.MaxIterations && ~PolesConverged
    
    % Update iteration counter
    iter = iter + 1;
    if ~Options.Silent
        fprintf('Iteration n.%d...\n',iter);
    end
    
    % Construction of various regressor matrix blocks
    [~,~,PHI,PSI] = ComputeBasis(y,kmax,t,u,realpoles,complpoles,0);
    
    % Here we use relaxed VF to rule out the trivial all-zero solution
    nonTrivialityConst = ones(1, size(PHI, 2)+1);
    nonTrivialityConst(1, 2:end) = mean(-PHI, 1);
    PHI = VFRelaxation(PHI, y);
    
    % Add free response terms (to take into account non-vanishing initial
    % conditions). This part would not be needed by "standard" TDVF applied
    % to LTI systems with vanishing initial conditions.
    PSI = updatePSIIntegrate(PSI, t, realpoles, complpoles);
    
    % Solution of the least squares system of equations. This QR operation
    % is not really needed, but it becomes necessary to extend to the MIMO
    % case efficiently. See references for details.
    curMat = [PSI,PHI];
    [~,R] = qr(curMat,0);
    Rsys = R(end-Order:end,end-Order:end);
    
    % Append nontriviality constraints and solve
    RsysNew  = [Rsys; nonTrivialityConst];
    bsys     = zeros(size(RsysNew, 1), 1);
    bsys(end) = 1; % RMS normalized
    residues = RsysNew\bsys;
    
    % Check convergence
    
    % Take the residues associated with the denominator
    WeightFunctRes(iter) = norm(residues(2:end));
    
    % Poles will be relocated one more time even if this criterion is met
    PolesConverged = (WeightFunctRes(iter) < Options.VFWeightNorm);
    
    % Pole relocation
    poles = PolesRelocation(realpoles,complpoles,residues(1),residues(2:end));
    polesOut(:,iter)=poles;
    
    % Classifies poles (a bit too conservative, but ok...)
    indrp = find(abs(imag(poles))<10*eps*abs(poles));
    indcp = find(imag(poles)>=10*eps*abs(poles));
    realpoles = real(poles(indrp)); %#ok
    complpoles = poles(indcp); %#ok
    
    if Options.debug
        fprintf(['\n Norm of Weighiting Function Coeffs:\n ' num2str(WeightFunctRes(iter)) '\n']);
        fprintf(['\n Poles:\n Real:' num2str(realpoles.') ' \n Complex: ' num2str(complpoles.') '\n']);
    end
    
    % Flip unstable poles into the left half plane if stability is enforced
    % (mandatory!)
    if Options.EnforceStability
        realpoles = -abs(realpoles);
        complpoles = -abs(real(complpoles))+1j*abs(imag(complpoles));
    end
    
    % Evaluate model residues at current iteration (if needed/required)
    if ~Options.SkipRes || iter == Options.MaxIterations || PolesConverged
        
        % Avoid computation of output poles convolution (not needed here)
        skipY = 1;
        [~,~,~,PSI] = ComputeBasis(y,kmax,t,u,realpoles,complpoles,skipY);
        
        % Add free response terms (handle initial conditions)
        nCoeffsSmallSignal = size(PSI, 2);
        PSI = updatePSIIntegrate(PSI, t, realpoles, complpoles);
        
        % Solve least squares system
        unknowns =PSI\y;
        coefficients = unknowns(1:nCoeffsSmallSignal, :, :);
        
        % Note: the coefficients taking into account non-vanishing
        % initial conditions are (see references)
        % coeffsInitialCond   = unknowns(nCoeffsSmallSignal+1:end, :, :);
        
        % Build (normalized) state-space realization of model        
        A=diag(realpoles);
        B=[ones(length(realpoles),1);kron(ones(length(complpoles),1),[2;0])];
        D=coefficients(1);
        C=coefficients(2:end)';
        for zz=1:length(complpoles)
            A=blkdiag(A,[real(complpoles(zz)),imag(complpoles(zz));-imag(complpoles(zz)),real(complpoles(zz))]);
        end
        Realization.A=A;
        Realization.B=B;
        Realization.C=C;
        Realization.D=D;
        
        % Finalize output arguments
        WeightFunctRes=WeightFunctRes(1:iter);
        polesOut=polesOut(:,1:iter);
            
    end
    
end

end

%% Internal function: time step extraction
function [dT,error] = extractTimeStep(t,tol)
%
% Retrieve time step (mean) and throw an error if sampling is nonuniform

if nargin<2
    tol = 1e-7;
end

dt = diff(t);
dtmax = max(dt);
dtmin = min(dt);
if (dtmax-dtmin)/mean(dt) > tol
    error = 'Time axis is not uniformly sampled';
else
    error = [];
end

dT = mean(dt);

end


%% Internal function: apply VF relaxation (append extra column)
function PHINew = VFRelaxation(PHI, y)

PHINew = [-y(:), PHI];

end

%% Internal function: time response of individual poles
function y = realPoleResp(poles, t)

y = exp(poles.*t);

end

%% Internal function: modify regressor matrix to account for nonvanishing initial conditions
function PSINew = updatePSIIntegrate(PSI, t, realpoles, complpoles)

nPoles = length(realpoles) + 2*length(complpoles);
PSINew = zeros(size(PSI, 1), size(PSI, 2)+ nPoles+1);

PSINew(:, 1:size(PSI, 2)) = PSI;

linInd = 1;

PSINew(:, size(PSI, 2)+linInd) = ones(length(t),1);
for iReal=1:length(realpoles)
    linInd = linInd +1;
    curPole = realpoles(iReal);
    PSINew(:, size(PSI, 2)+linInd) = (realPoleResp(curPole, t) - 1)./curPole;
end

ind1 = size(PSI, 2)+linInd-1;
for iComplx=1:length(complpoles)
    
    curPole = complpoles(iComplx);
    ind1 = ind1 +2;
    ind2 = ind1+1;
    PSINew(:, ind1:ind2) =  (cmplxPolesRespInteg(curPole, t));
end

end

%% Internal function: compute response of complex poles enforcing realness
function y = cmplxPolesRespInteg(poles, t)

y = zeros(length(t), 2);
resp =(exp(poles.*t)-1)./poles;
y(:, 1) = 2*real(resp);
y(:, 2) = 2*imag(resp);

end

%% Internal function: assemble the various VF regressor matrix blocks
function [Ask,bsk,PHINew,PSINew] = ComputeBasis(y,kmax,t,u,realpoles,complpoles,skipY)
% A Function to compute time domain basis functions for the TDVF code. This
% is not optimized since the convolutions are computed with the matlab
% built-in function LSIM. Much more efficient implementations are possible,
% e.g. through recursive convolutions applied throguh IIR filters.
%
% Author: Tommaso Bradde
% EMC group, Politecnico di Torino, Italy
% March 2023

nRealPoles = length(realpoles);
nComplPoles = length(complpoles);

pin = size(u,2);
pout = size(y,2);

PSINew = zeros(kmax,(1+nRealPoles+2*nComplPoles)*pin );

CurIndex = 0;
for iport = 1:pin
    
    CurIndex = CurIndex + 1;
    CurU = u(:,iport);
    PSINew(:,CurIndex)=CurU;
    
    for ipole = 1:nRealPoles
        
        curPole = realpoles(ipole);
        CurIndex = CurIndex +1;
        PSINew(:,CurIndex)=lsim(ss(curPole,1,1,0),CurU,t);
    end
    
    for ipole = 1:nComplPoles
        curPole = complpoles(ipole);
        
        CurIndex = CurIndex +1;
        
        Ypos=lsim(ss(curPole,1,1,0),CurU,t);
        Yneg=lsim(ss(conj(curPole),1,1,0),CurU,t);
        
        col1=Ypos+Yneg;
        col2=1i*Ypos-1i*Yneg;
        
        PSINew(:,CurIndex:CurIndex+1)=[col1,col2];
        
        CurIndex = CurIndex +1;
        
    end
    
end

%% Output/pole convolution
if ~skipY
    
    PHINew =zeros(kmax*pout,nRealPoles+2*nComplPoles);
    
    for iport = 1:pout
        CurIndex = (iport-1)*kmax+1;
        CurY = y(:,iport);
        PolesIndex=1;
        for ipole = 1:nRealPoles
            curPole = realpoles(ipole);
            
            PHINew(CurIndex:CurIndex+(kmax-1),PolesIndex) = lsim(ss(curPole,1,1,0),CurY,t);
            PolesIndex=PolesIndex+1;
            
        end
        
        PolesIndex = nRealPoles+1;
        
        for ipole = 1:nComplPoles
            curPole = complpoles(ipole);
            
            Ypos=lsim(ss(curPole,1,1,0),CurY,t);
            Yneg=lsim(ss(conj(curPole),1,1,0),CurY,t);
            
            col1=Ypos+Yneg;
            col2=1i*Ypos-1i*Yneg;
            
            PHINew(CurIndex:CurIndex+(kmax-1),PolesIndex:PolesIndex+1)=[col1,col2];
            PolesIndex = PolesIndex + 2;
        end
        
    end
    
else
    PHINew=[];
end

PHINew=-PHINew;

% Build regressor matrix
if ~skipY
    Ask = [PHINew,kron(eye(pout),PSINew)];
    bsk = y(:);
else
    Ask = kron(eye(pout),PSINew);
    bsk = y(:);
end

end

%% Internal function: compute new poles
function poles = PolesRelocation(realpoles,complpoles,asympt0,residues)

A=diag(realpoles);
B=[ones(length(realpoles),1);kron(ones(length(complpoles),1),[2;0])];D=asympt0;
C=residues';
for zz=1:length(complpoles)
    % Inefficient but more readable...
    A=blkdiag(A,[real(complpoles(zz)),imag(complpoles(zz));-imag(complpoles(zz)),real(complpoles(zz))]);
end
poles=eig(A-B*C/D);

end